"""
Code for `sarif csv` command.
"""

import csv
import os

from sarif import sarif_file
from sarif.sarif_file import SarifFileSet
from sarif.sarif_file_utils import combine_record_code_and_description


def generate_csv(input_files: SarifFileSet, output: str, output_multiple_files: bool):
    """
    Generate a CSV file containing the list of issues from the SARIF files.
    sarif_dict is a dict from filename to deserialized SARIF data.
    """
    output_file = output
    if output_multiple_files:
        for input_file in input_files:
            output_file_name = input_file.get_file_name_without_extension() + ".csv"
            print(
                "Writing CSV summary of",
                input_file.get_file_name(),
                "to",
                output_file_name,
            )
            _write_to_csv(input_file, os.path.join(output, output_file_name))
            filter_stats = input_file.get_filter_stats()
            if filter_stats:
                print(f"  Results are filtered by {filter_stats}")
        output_file = os.path.join(output, "static_analysis_output.csv")
    source_description = input_files.get_description()
    print(
        "Writing CSV summary for",
        source_description,
        "to",
        os.path.basename(output_file),
    )
    _write_to_csv(input_files, output_file)
    filter_stats = input_files.get_filter_stats()
    if filter_stats:
        print(f"  Results are filtered by {filter_stats}")


def _write_to_csv(file_or_files, output_file):
    """
    Write out the errors to a CSV file so that a human can do further analysis.
    """
    list_of_errors = file_or_files.get_records()
    severities = file_or_files.get_severities()
    # newline="" to avoid \r\r\n - see https://stackoverflow.com/a/3191811/316578
    with open(output_file, "w", encoding="utf-8", newline="") as file_out:
        writer = csv.DictWriter(
            file_out, sarif_file.get_record_headings(file_or_files.has_blame_info())
        )
        writer.writeheader()
        for severity in severities:
            errors_of_severity = [
                e for e in list_of_errors if e["Severity"] == severity
            ]
            sorted_errors_by_severity = sorted(
                errors_of_severity, key=combine_record_code_and_description
            )
            writer.writerows(error_dict for error_dict in sorted_errors_by_severity)



"""
Program entry point for sarif-tools on the command line.
"""

import argparse
import os
import sys

from sarif import loader, sarif_file, __version__ as SARIF_TOOLS_PACKAGE_VERSION
from sarif.filter.general_filter import load_filter_file

from sarif.operations import (
    blame_op,
    codeclimate_op,
    copy_op,
    csv_op,
    diff_op,
    html_op,
    emacs_op,
    info_op,
    ls_op,
    summary_op,
    trend_op,
    upgrade_filter_op,
    word_op,
)


def main():
    """
    Entry point function.
    """
    args, unknown_args = ARG_PARSER.parse_known_args()

    if args.debug:
        _print_version()
        print(f"Running code from {__file__}")
        known_args_summary = ", ".join(
            f"{key}={getattr(args, key)}" for key in vars(args)
        )
        print(f"Known arguments: {known_args_summary}")
        if args.version:
            return 0
    elif args.version:
        _print_version()
        return 0

    if unknown_args:
        if any(
            unknown_arg.startswith("--blame-filter")
            or unknown_arg.startswith("-b=")
            or unknown_arg == "-b"
            for unknown_arg in unknown_args
        ):
            print("ERROR: --blame-filter was removed in v2.0.0.")
            print(
                "Run the upgrade-filter command to convert your blame filter to the new filter format, then pass via --filter option."
            )
        args = ARG_PARSER.parse_args()

    exitcode = args.func(args)
    return exitcode


def _create_arg_parser():
    cmd_list = "commands:\n"
    max_cmd_length = max(len(cmd) for cmd in _COMMANDS)
    col_width = max_cmd_length + 2
    for cmd, cmd_attributes in _COMMANDS.items():
        cmd_list += cmd.ljust(col_width) + cmd_attributes["desc"] + "\n"
    cmd_list += "Run `sarif <COMMAND> --help` for command-specific help."
    parser = argparse.ArgumentParser(
        prog="sarif",
        description="Process sets of SARIF files",
        epilog=cmd_list,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.set_defaults(func=_usage_command)
    subparsers = parser.add_subparsers(dest="command", help="command")
    subparser = {}
    for cmd, cmd_attributes in _COMMANDS.items():
        subparser[cmd] = subparsers.add_parser(cmd, description=cmd_attributes["desc"])
        subparser[cmd].set_defaults(func=cmd_attributes["fn"])

    # Common options
    parser.add_argument("--version", "-v", action="store_true")
    parser.add_argument(
        "--debug", action="store_true", help="Print information useful for debugging"
    )
    parser.add_argument(
        "--check",
        "-x",
        type=str,
        choices=sarif_file.SARIF_SEVERITIES_WITH_NONE,
        help="Exit with error code if there are any issues of the specified level "
        + "(or for diff, an increase in issues at that level).",
    )

    for cmd in [
        "blame",
        "codeclimate",
        "csv",
        "html",
        "emacs",
        "summary",
        "word",
        "upgrade-filter",
    ]:
        subparser[cmd].add_argument(
            "--output", "-o", type=str, metavar="PATH", help="Output file or directory"
        )
    for cmd in ["copy", "diff", "info", "ls", "trend", "usage"]:
        subparser[cmd].add_argument(
            "--output", "-o", type=str, metavar="FILE", help="Output file"
        )

    for cmd in [
        "codeclimate",
        "copy",
        "csv",
        "diff",
        "summary",
        "html",
        "emacs",
        "trend",
        "word",
    ]:
        subparser[cmd].add_argument(
            "--filter",
            "-b",
            type=str,
            metavar="FILE",
            help="Specify the filter file to apply.  See README for format.",
        )

    # Command-specific options
    subparser["blame"].add_argument(
        "--code",
        "-c",
        metavar="PATH",
        type=str,
        help="Path to git repository; if not specified, the current working directory is used",
    )
    subparser["copy"].add_argument(
        "--timestamp",
        "-t",
        action="store_true",
        help='Append current timestamp to output filename in the "yyyymmddThhmmssZ" format used by '
        "the `sarif trend` command",
    )
    # codeclimate and csv default to no trimming
    for cmd in ["codeclimate", "csv"]:
        subparser[cmd].add_argument(
            "--autotrim",
            "-a",
            action="store_true",
            help="Strip off the common prefix of paths in the CSV output",
        )
    # word and html default to trimming
    for cmd in ["html", "emacs", "word"]:
        subparser[cmd].add_argument(
            "--no-autotrim",
            "-n",
            action="store_true",
            help="Do not strip off the common prefix of paths in the output document",
        )
        subparser[cmd].add_argument(
            "--image",
            type=str,
            help="Image to include at top of file - SARIF logo by default",
        )
    # codeclimate, csv, html and word allow trimmable paths to be specified
    for cmd in ["codeclimate", "csv", "word", "html", "emacs"]:
        subparser[cmd].add_argument(
            "--trim",
            metavar="PREFIX",
            action="append",
            type=str,
            help="Prefix to strip from issue paths, e.g. the checkout directory on the build agent",
        )
    # Most commands take an arbitrary list of SARIF files or directories
    for cmd in _COMMANDS:
        if cmd not in ["diff", "upgrade-filter", "usage", "version"]:
            subparser[cmd].add_argument(
                "files_or_dirs",
                metavar="file_or_dir",
                type=str,
                nargs="*",
                default=["."],
                help="A SARIF file or a directory containing SARIF files",
            )
    subparser["diff"].add_argument(
        "old_file_or_dir",
        type=str,
        nargs=1,
        help="An old SARIF file or a directory containing the old SARIF files",
    )
    subparser["diff"].add_argument(
        "new_file_or_dir",
        type=str,
        nargs=1,
        help="A new SARIF file or a directory containing the new SARIF files",
    )

    subparser["trend"].add_argument(
        "--dateformat",
        "-f",
        type=str,
        choices=["dmy", "mdy", "ymd"],
        default="dmy",
        help="Date component order to use in output CSV.  Default is `dmy`",
    )

    subparser["upgrade-filter"].add_argument(
        "files_or_dirs",
        metavar="file",
        type=str,
        nargs="*",
        default=["."],
        help="A v1-style blame-filter file",
    )

    return parser


def _check(input_files: sarif_file.SarifFileSet, check_level):
    ret = 0
    if check_level:
        counts = input_files.get_result_count_by_severity()
        for severity in sarif_file.SARIF_SEVERITIES_WITH_NONE:
            ret += counts.get(severity, 0)
            if severity == check_level:
                break
    if ret > 0:
        sys.stderr.write(
            f"Check: exiting with return code {ret} due to issues at or above {check_level} "
            "severity\n"
        )
    return ret


def _init_filtering(input_files, args):
    if args.filter:
        filters = load_filter_file(args.filter)
        input_files.init_general_filter(*filters)


def _init_path_prefix_stripping(input_files, args, strip_by_default):
    if strip_by_default:
        autotrim = not args.no_autotrim
    else:
        autotrim = args.autotrim
    trim_paths = args.trim
    if autotrim or trim_paths:
        input_files.init_path_prefix_stripping(autotrim, trim_paths)


def _ensure_dir(dir_path):
    """
    Create directory if it does not exist
    """
    if dir_path and not os.path.isdir(dir_path):
        os.makedirs(dir_path)


def _prepare_output(
    input_files: sarif_file.SarifFileSet, output_arg, output_file_extension: str
):
    """
    Returns (output, output_multiple_files)
    output is args.output, or if that wasn't specified, a default output file based on the inputs
    and the file extension.
    output_multiple_files determines whether to output one file per input plus a totals file.
    It is false if there is only one input file, or args.output is a file that exists,
    or args.output ends with the expected file extension.
    """
    input_file_count = len(input_files)
    if input_file_count == 0:
        return ("static_analysis_output" + output_file_extension, False)
    if input_file_count == 1:
        derived_output_filename = (
            input_files[0].get_file_name_without_extension() + output_file_extension
        )
        if output_arg:
            if os.path.isdir(output_arg):
                return (os.path.join(output_arg, derived_output_filename), False)
            _ensure_dir(os.path.dirname(output_arg))
            return (output_arg, False)
        return (derived_output_filename, False)
    # Multiple input files
    if output_arg:
        if os.path.isfile(output_arg) or output_arg.strip().upper().endswith(
            output_file_extension.upper()
        ):
            # Output single file, even though there are multiple input files.
            _ensure_dir(os.path.dirname(output_arg))
            return (output_arg, False)
        _ensure_dir(output_arg)
        return (output_arg, True)
    return (os.getcwd(), True)


####################################### Command handlers #######################################


def _blame_command(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    (output, multiple_file_output) = _prepare_output(input_files, args.output, ".sarif")
    blame_op.enhance_with_blame(
        input_files, args.code or os.getcwd(), output, multiple_file_output
    )
    return _check(input_files, args.check)


def _codeclimate_command(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    input_files.init_default_line_number_1()
    _init_path_prefix_stripping(input_files, args, strip_by_default=False)
    _init_filtering(input_files, args)
    (output, multiple_file_output) = _prepare_output(input_files, args.output, ".json")
    codeclimate_op.generate(input_files, output, multiple_file_output)
    return _check(input_files, args.check)


def _copy_command(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    _init_filtering(input_files, args)
    output = args.output or "out.sarif"
    output_sarif_file_set = copy_op.generate_sarif(
        input_files,
        output,
        args.timestamp,
        SARIF_TOOLS_PACKAGE_VERSION,
        " ".join(sys.argv),
    )
    return _check(output_sarif_file_set, args.check)


def _csv_command(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    input_files.init_default_line_number_1()
    _init_path_prefix_stripping(input_files, args, strip_by_default=False)
    _init_filtering(input_files, args)
    (output, multiple_file_output) = _prepare_output(input_files, args.output, ".csv")
    csv_op.generate_csv(input_files, output, multiple_file_output)
    return _check(input_files, args.check)


def _diff_command(args):
    old_sarif = loader.load_sarif_files(args.old_file_or_dir[0])
    new_sarif = loader.load_sarif_files(args.new_file_or_dir[0])
    _init_filtering(old_sarif, args)
    _init_filtering(new_sarif, args)
    return diff_op.print_diff(old_sarif, new_sarif, args.output, args.check)


def _html_command(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    input_files.init_default_line_number_1()
    _init_path_prefix_stripping(input_files, args, strip_by_default=True)
    _init_filtering(input_files, args)
    (output, multiple_file_output) = _prepare_output(input_files, args.output, ".html")
    html_op.generate_html(input_files, args.image, output, multiple_file_output)
    return _check(input_files, args.check)


def _emacs_command(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    input_files.init_default_line_number_1()
    _init_path_prefix_stripping(input_files, args, strip_by_default=True)
    _init_filtering(input_files, args)
    (output, multiple_file_output) = _prepare_output(input_files, args.output, ".txt")
    emacs_op.generate_compile(input_files, output, multiple_file_output)
    return _check(input_files, args.check)


def _info_command(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    info_op.generate_info(input_files, args.output)
    if args.check:
        return _check(input_files, args.check)
    return 0


def _ls_command(args):
    ls_op.print_ls(args.files_or_dirs, args.output)
    if args.check:
        input_files = loader.load_sarif_files(*args.files_or_dirs)
        return _check(input_files, args.check)
    return 0


def _summary_command(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    _init_filtering(input_files, args)
    (output, multiple_file_output) = (None, False)
    if args.output:
        (output, multiple_file_output) = _prepare_output(
            input_files, args.output, ".txt"
        )
    summary_op.generate_summary(input_files, output, multiple_file_output)
    return _check(input_files, args.check)


def _trend_command(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    input_files.init_default_line_number_1()
    _init_filtering(input_files, args)
    if args.output:
        _ensure_dir(os.path.dirname(args.output))
        output = args.output
    else:
        output = "static_analysis_trend.csv"
    trend_op.generate_trend_csv(input_files, output, args.dateformat)
    return _check(input_files, args.check)


def _upgrade_filter_command(args):
    old_filter_files = args.files_or_dirs
    single_output_file = None
    output_dir = None
    if len(old_filter_files) == 1:
        if args.output and os.path.isdir(args.output):
            output_dir = args.output
        else:
            single_output_file = args.output or old_filter_files[0] + ".yaml"
    elif args.output:
        output_dir = args.output
    else:
        output_dir = os.path.dirname(args.output)
    for old_filter_file in old_filter_files:
        output_file = single_output_file or os.path.join(
            output_dir, os.path.basename(old_filter_file) + ".yaml"
        )
        upgrade_filter_op.upgrade_filter_file(old_filter_file, output_file)
    return 0


def _usage_command(args):
    if hasattr(args, "output") and args.output:
        with open(args.output, "w", encoding="utf-8") as file_out:
            ARG_PARSER.print_help(file_out)
        print("Wrote usage instructions to", args.output)
    else:
        ARG_PARSER.print_help()
    if args.check:
        sys.stderr.write("Spurious --check argument")
        return 1
    return 0


def _version_command(args):
    _print_version(not args.version)


def _print_version(bare=False):
    print(
        SARIF_TOOLS_PACKAGE_VERSION
        if bare
        else f"SARIF tools v{SARIF_TOOLS_PACKAGE_VERSION}"
    )


def _word_command(args):
    input_files = loader.load_sarif_files(*args.files_or_dirs)
    input_files.init_default_line_number_1()
    _init_path_prefix_stripping(input_files, args, strip_by_default=True)
    _init_filtering(input_files, args)
    (output, multiple_file_output) = _prepare_output(input_files, args.output, ".docx")
    word_op.generate_word_docs_from_sarif_inputs(
        input_files, args.image, output, multiple_file_output
    )
    return _check(input_files, args.check)


_COMMANDS = {
    "blame": {
        "fn": _blame_command,
        "desc": "Enhance SARIF file with information from `git blame`",
    },
    "codeclimate": {
        "fn": _codeclimate_command,
        "desc": "Write a JSON representation in Code Climate format of SARIF file(s) "
        "for viewing as a Code Quality report in GitLab UI",
    },
    "copy": {
        "fn": _copy_command,
        "desc": "Write a new SARIF file containing optionally-filtered data from other SARIF file(s)",
    },
    "csv": {
        "fn": _csv_command,
        "desc": "Write a CSV file listing the issues from the SARIF files(s) specified",
    },
    "diff": {
        "fn": _diff_command,
        "desc": "Find the difference between two [sets of] SARIF files",
    },
    "emacs": {
        "fn": _emacs_command,
        "desc": "Write a representation of SARIF file(s) for viewing in emacs",
    },
    "html": {
        "fn": _html_command,
        "desc": "Write an HTML representation of SARIF file(s) for viewing in a web browser",
    },
    "info": {
        "fn": _info_command,
        "desc": "Print information about SARIF file(s) structure",
    },
    "ls": {
        "fn": _ls_command,
        "desc": "List all SARIF files in the directories specified",
    },
    "summary": {
        "fn": _summary_command,
        "desc": "Write a text summary with the counts of issues from the SARIF files(s) specified",
    },
    "trend": {
        "fn": _trend_command,
        "desc": "Write a CSV file with time series data from SARIF files with "
        '"yyyymmddThhmmssZ" timestamps in their filenames',
    },
    "upgrade-filter": {
        "fn": _upgrade_filter_command,
        "desc": "Upgrade a sarif-tools v1-style blame filter file to a v2-style filter YAML file",
    },
    "usage": {
        "fn": _usage_command,
        "desc": "(Command optional) - print usage and exit",
    },
    "version": {"fn": _version_command, "desc": "Print version and exit"},
    "word": {
        "fn": _word_command,
        "desc": "Produce MS Word .docx summaries of the SARIF files specified",
    },
}

ARG_PARSER = _create_arg_parser()



"""
Functions for generating charts from SARIF data
"""

import io
import matplotlib.pyplot as plt


def generate_severity_pie_chart(report, output_file=None):
    """
    Generate a pie chart from the breakdown of issues by severity.
    The slices are ordered and plotted counter-clockwise.  The return
    value is truthy if the number of issues is not zero, False otherwise.
    If `output_file` is `None`, return the bytes of the pie chart image in
    png format.  Otherwise, write the bytes to the file specified (image
    format inferred from filename).
    """
    sizes = []
    labels = []
    explode = []
    for severity in report.get_severities():
        count = report.get_issue_count_for_severity(severity)
        if count > 0:
            sizes.append(count)
            labels.append(severity)
            explode.append(0.1)  # could add more logic to highlight specific severities

    any_issues = bool(sizes)
    if any_issues:
        _fig1, ax1 = plt.subplots()
        ax1.pie(
            sizes,
            explode=explode,
            labels=labels,
            autopct="%1.1f%%",
            shadow=True,
            startangle=90,
        )
        ax1.axis("equal")

        if output_file:
            plt.savefig(output_file)
        else:
            byte_buffer = io.BytesIO()
            plt.savefig(byte_buffer, format="png")
            return byte_buffer.getbuffer()
    return any_issues



"""
Code for `sarif trend` command.
"""

import csv
from typing import Dict, List

from sarif import sarif_file
from sarif.sarif_file import SarifFileSet

TIMESTAMP_COLUMNS = ["Date", "Tool", *sarif_file.SARIF_SEVERITIES_WITH_NONE]


def generate_trend_csv(input_files: SarifFileSet, output_file: str, dateformat: str):
    """
    Generate a timeline csv of the issues from the SARIF files.  Each SARIF file must contain a
    timestamp of the form 20211012T110000Z in its filename.
    sarif_dict is a dict from filename to deserialized SARIF data.
    output_file is the name of a CSV file to write, or if None, the name
    `static_analysis_trend.csv` will be used.
    """
    if not output_file:
        output_file = "static_analysis_trend.csv"

    error_storage = []
    for input_file in input_files:
        input_file_name = input_file.get_file_name()
        print("Processing", input_file_name)
        error_list = input_file.get_records()
        tool_name = "/".join(input_file.get_distinct_tool_names())
        # Date parsing
        parsed_date = input_file.get_filename_timestamp()
        if not parsed_date:
            raise ValueError(f"Unable to parse date from filename: {input_file_name}")

        # Turn the date into something that looks nice in excel (d/m/y UK date format)
        dstr = parsed_date[0]
        (year, month, day, hour, minute) = (
            dstr[0:4],
            dstr[4:6],
            dstr[6:8],
            dstr[9:11],
            dstr[11:13],
        )
        if dateformat == "ymd":
            excel_date = f"{year}-{month}-{day} {hour}:{minute}"
        elif dateformat == "mdy":
            excel_date = f"{month}/{day}/{year} {hour}:{minute}"
        else:
            excel_date = f"{day}/{month}/{year} {hour}:{minute}"

        # Store data
        error_storage.append(
            _store_errors(parsed_date, excel_date, tool_name, error_list)
        )

    error_storage.sort(key=lambda record: record["_timestamp"])

    print("Writing trend CSV to", output_file)
    _write_csv(output_file, error_storage)
    filter_stats = input_files.get_filter_stats()
    if filter_stats:
        print(f"  Results are filtered by {filter_stats}")


def _write_csv(output_file: str, error_storage: List[Dict]) -> None:
    # newline="" to avoid \r\r\n - see https://stackoverflow.com/a/3191811/316578
    with open(output_file, "w", newline="", encoding="utf-8") as file_out:
        writer = csv.DictWriter(file_out, TIMESTAMP_COLUMNS, extrasaction="ignore")
        writer.writeheader()
        for key in error_storage:
            writer.writerow(key)


def _store_errors(timestamp, excel_date, tool: str, list_of_errors: List[Dict]) -> Dict:
    results = {
        "_timestamp": timestamp,  # not written to CSV, but used for sorting
        "Date": excel_date,
        "Tool": tool,
    }
    for severity in sarif_file.SARIF_SEVERITIES_WITH_NONE:
        error_count = sum(1 for e in list_of_errors if severity in e["Severity"])
        results[severity] = error_count

    return results



"""
Reusable utility functions for handling the SARIF format.

Primarily interrogating the `result` JSON defined at
https://docs.oasis-open.org/sarif/sarif/v2.1.0/cs01/sarif-v2.1.0-cs01.html#_Toc16012594
"""

import textwrap
from typing import Tuple

# SARIF severity levels as per
# https://docs.oasis-open.org/sarif/sarif/v2.1.0/sarif-v2.1.0.html#_Toc141790898
SARIF_SEVERITIES_WITHOUT_NONE = ["error", "warning", "note"]
SARIF_SEVERITIES_WITH_NONE = SARIF_SEVERITIES_WITHOUT_NONE + ["none"]


def combine_code_and_description(code: str, description: str) -> str:
    """
    Combine code and description into one string, keeping total length under 120 characters.
    """
    length_budget = 120
    if code:
        code = code.strip()
        length_budget -= len(code) + 1  # Allow issue code and space character
    continuation_placeholder = " ..."
    # Allow extra space when truncating for continuation characters
    length_budget_pre_continuation = length_budget - len(continuation_placeholder)
    if length_budget_pre_continuation < 10:
        # Don't include description if it would be very short due to long code
        return code
    if description:
        if "\n" in description:
            description = description[: description.index("\n")]
        if description.startswith(code):
            # Don't duplicate the code
            description = description[len(code) :]
        description = description.strip()
    if description:
        if len(description) > length_budget:
            shorter_description = textwrap.shorten(
                description,
                width=length_budget_pre_continuation,
                placeholder=continuation_placeholder,
            )
            if len(shorter_description) < length_budget_pre_continuation - 40:
                # Word wrap shortens the description significantly, so truncate mid-word instead
                description = (
                    description[:length_budget_pre_continuation]
                    + continuation_placeholder
                )
            else:
                description = shorter_description
        if code:
            return f"{code.strip()} {description}"
        return description
    if code:
        return code
    return "<NONE>"


def combine_record_code_and_description(record: dict) -> str:
    """
    Combine code and description fields into one string.
    """
    return combine_code_and_description(record["Code"], record["Description"])


def read_result_location(result) -> Tuple[str, str]:
    """
    Extract the file path and line number strings from the Result.

    Tools store this in different ways, so this function tries a few different JSON locations.
    """
    file_path = None
    line_number = None
    locations = result.get("locations", [])
    if locations and isinstance(locations, list):
        location = locations[0]
        physical_location = location.get("physicalLocation", {})
        # SpotBugs has some errors with no line number so deal with them by just leaving it at 1
        line_number = physical_location.get("region", {}).get("startLine", None)
        # For file name, first try the location written by DevSkim
        file_path = (
            location.get("physicalLocation", {})
            .get("address", {})
            .get("fullyQualifiedName", None)
        )
        if not file_path:
            # Next try the physical location written by MobSF and by SpotBugs (for some errors)
            file_path = (
                location.get("physicalLocation", {})
                .get("artifactLocation", {})
                .get("uri", None)
            )
        if not file_path:
            logical_locations = location.get("logicalLocations", None)
            if logical_locations:
                # Finally, try the logical location written by SpotBugs for some errors
                file_path = logical_locations[0].get("fullyQualifiedName", None)
    return (file_path, line_number)


def record_sort_key(record: dict) -> str:
    """Get a sort key for the record."""
    return (
        combine_record_code_and_description(record)
        + record["Location"]
        + str(record["Line"]).zfill(6)
    )



"""
Code for `sarif html` command.
"""

import base64
from datetime import datetime
import os

from jinja2 import Environment, FileSystemLoader, select_autoescape

from sarif import charts
from sarif import sarif_file

_THIS_MODULE_PATH = os.path.dirname(__file__)

_TEMPLATES_PATH = os.path.join(_THIS_MODULE_PATH, "templates")

_ENV = Environment(
    loader=FileSystemLoader(searchpath=_TEMPLATES_PATH),
    autoescape=select_autoescape(),
)


def generate_html(
    input_files: sarif_file.SarifFileSet,
    image_file: str,
    output: str,
    output_multiple_files: bool,
):
    """
    Generate HTML file from the input files.
    """
    date_val = datetime.now()

    if image_file:
        image_mime_type = "image/" + os.path.splitext(image_file)[-1]
        if image_mime_type == "image/jpg":
            image_mime_type = "image/jpeg"
        with open(image_file, "rb") as input_file:
            image_data = input_file.read()

        image_data_base64 = base64.b64encode(image_data).decode("utf-8")
    else:
        image_mime_type = None
        image_data_base64 = None

    output_file = output
    if output_multiple_files:
        for input_file in input_files:
            output_file_name = input_file.get_file_name_without_extension() + ".html"
            print(
                "Writing HTML report for",
                input_file.get_file_name(),
                "to",
                output_file_name,
            )
            _generate_single_html(
                input_file,
                os.path.join(output, output_file_name),
                date_val,
                image_mime_type,
                image_data_base64,
            )
        output_file = os.path.join(output, "static_analysis_output.html")
    source_description = input_files.get_description()
    print(
        "Writing HTML report for",
        source_description,
        "to",
        os.path.basename(output_file),
    )
    _generate_single_html(
        input_files, output_file, date_val, image_mime_type, image_data_base64
    )


def _generate_single_html(
    input_file, output_file, date_val, image_mime_type, image_data_base64
):
    all_tools = input_file.get_distinct_tool_names()
    report = input_file.get_report()

    total_distinct_issue_codes = 0
    problems = []
    severities = report.get_severities()

    for severity in severities:
        distinct_issue_codes = report.get_issue_type_count_for_severity(severity)

        total_distinct_issue_codes += distinct_issue_codes

        severity_details = _enrich_details(
            report.get_issues_grouped_by_type_for_severity(severity)
        )

        severity_section = {
            "type": severity,
            "count": distinct_issue_codes,
            "details": severity_details,
        }

        problems.append(severity_section)

    chart_data = charts.generate_severity_pie_chart(report, output_file=None)
    if chart_data:
        chart_image_data_base64 = base64.b64encode(chart_data).decode("utf-8")
    else:
        chart_image_data_base64 = None

    filtered = None
    filter_stats = input_file.get_filter_stats()
    if filter_stats:
        filtered = f"Results were filtered by {filter_stats}."

    template = _ENV.get_template("sarif_summary.html")
    html_content = template.render(
        report_type=", ".join(all_tools),
        report_date=date_val,
        severities=", ".join(severities),
        total=total_distinct_issue_codes,
        problems=problems,
        image_mime_type=image_mime_type,
        image_data_base64=image_data_base64,
        chart_image_data_base64=chart_image_data_base64,
        filtered=filtered,
    )

    with open(output_file, "wt", encoding="utf-8") as file_out:
        file_out.write(html_content)


def _enrich_details(records_of_severity):
    return [
        {"code": key, "count": len(records), "details": records}
        for (key, records) in records_of_severity.items()
    ]



"""
Generate summary of given JSON or given JSON Document in to Microsoft Word Documents.
This functionality uses a python module called

python-docx - a Python library for creating and updating Microsoft Word (.docx) files.

https://python-docx.readthedocs.io/

"""

from datetime import datetime
import os

import docx
from docx import oxml
from docx import shared
from docx.enum import text
from docx.oxml import ns

from sarif import charts, sarif_file
from sarif.sarif_file_utils import combine_record_code_and_description


def generate_word_docs_from_sarif_inputs(
    input_files: sarif_file.SarifFileSet,
    image_file: str,
    output: str,
    output_multiple_files: bool,
):
    """
    Convert SARIF input to Word file output.
    """
    if not input_files:
        raise ValueError("No input files specified!")

    output_file = output
    output_file_name = output
    if output_multiple_files:
        for input_file in input_files:
            output_file_name = input_file.get_file_name_without_extension() + ".docx"
            print(
                "Writing Word summary of",
                input_file.get_file_name(),
                "to",
                output_file_name,
            )
            report = input_file.get_report()
            _generate_word_summary(
                input_file,
                report,
                os.path.join(output, output_file_name),
                image_file,
            )
        output_file_name = "static_analysis_output.docx"
        output_file = os.path.join(output, output_file_name)

    source_description = input_files.get_description()
    print("Writing Word summary of", source_description, "to", output_file_name)
    report = input_files.get_report()
    _generate_word_summary(input_files, report, output_file, image_file)


def _generate_word_summary(sarif_data, report, output_file, image_file):
    # Create a new document
    document = docx.Document()

    severities = report.get_severities()
    _add_heading_and_highlevel_info(
        document, sarif_data, report, severities, output_file, image_file
    )
    _dump_errors_summary_by_sev(document, report, severities)
    _dump_each_error_in_detail(document, report, severities)

    # finally, save the document.
    document.save(output_file)


def _add_heading_and_highlevel_info(
    document, sarif_data, report, severities, output_file, image_path
):
    tool_name = ", ".join(sarif_data.get_distinct_tool_names())
    heading = f"Sarif Summary: {tool_name}"

    if image_path:
        document.add_picture(image_path)
        last_paragraph = document.paragraphs[-1]
        last_paragraph.alignment = text.WD_PARAGRAPH_ALIGNMENT.CENTER

    document.add_heading(heading, 0)
    document.add_paragraph(f"Document generated on: {datetime.now()}")

    sevs = ", ".join(severities)
    document.add_paragraph(
        f"Total number of various severities ({sevs}): {sarif_data.get_result_count()}"
    )
    filter_stats = sarif_data.get_filter_stats()
    if filter_stats:
        document.add_paragraph(f"Results were filtered by {filter_stats}.")

    pie_chart_image_file_path = output_file.replace(".docx", "_severity_pie_chart.png")
    if charts.generate_severity_pie_chart(report, pie_chart_image_file_path):
        document.add_picture(pie_chart_image_file_path)
    last_paragraph = document.paragraphs[-1]
    last_paragraph.alignment = text.WD_PARAGRAPH_ALIGNMENT.CENTER

    document.add_page_break()


def _dump_errors_summary_by_sev(document, report, severities):
    """
    For each severity level (in priority order): create a list of the errors of
    that severity, print out how many there are and then do some further analysis
    of which error codes are present.
    """
    for severity in severities:
        errors_of_severity = report.get_issue_type_count_for_severity(severity)
        document.add_heading(f"Severity : {severity} [ {errors_of_severity} ]", level=1)
        sorted_dict = report.get_issue_type_histogram_for_severity(severity)
        if sorted_dict:
            for key, count in sorted_dict.items():
                document.add_paragraph(f"{key}: {count}", style="List Bullet")
        else:
            document.add_paragraph("None", style="List Bullet")


def _dump_each_error_in_detail(document, report, severities):
    """
    Write out the errors to a table so that a human can do further analysis.
    """
    document.add_page_break()

    for severity in severities:
        errors_of_severity = report.get_issues_for_severity(severity)
        # Sample:
        # [{'Location': 'C:\\Max\\AccessionAndroid\\scripts\\parse_coverage.py', 'Line': 119,
        #       'Severity': 'error', 'Code': 'DS126186 Disabled certificate validation'},
        # {'Location': 'C:\\Max\\AccessionAndroid\\scripts\\parse_code_stats.py', 'Line': 61,
        #       'Severity': 'error', 'Code': 'DS126186 Disabled certificate validation'},
        # ]
        if errors_of_severity:
            document.add_heading(f"Severity : {severity}", level=2)
            table = document.add_table(rows=1 + len(errors_of_severity), cols=3)

            table.style = "Table Grid"  # ColorfulGrid-Accent5'
            table.autofit = False

            table.alignment = text.WD_TAB_ALIGNMENT.CENTER

            # Cell widths
            widths = [shared.Inches(2), shared.Inches(4), shared.Inches(0.5)]

            # To avoid performance problems with large tables, prepare the entries first in this
            # list, then iterate the table cells and copy them in.
            # First populate the header row
            cells_text = ["Code", "Location", "Line"]

            hdr_cells = table.rows[0].cells
            for i in range(3):
                table.rows[0].cells[i]._tc.get_or_add_tcPr().append(
                    oxml.parse_xml(
                        r'<w:shd {} w:fill="5fe3d8"/>'.format(ns.nsdecls("w"))
                    )
                )
                run = hdr_cells[i].paragraphs[0].add_run(cells_text[i])
                run.bold = True
                hdr_cells[i].paragraphs[
                    0
                ].alignment = text.WD_PARAGRAPH_ALIGNMENT.CENTER
                hdr_cells[i].width = widths[i]

            for eachrow in errors_of_severity:
                cells_text += [
                    combine_record_code_and_description(eachrow),
                    eachrow["Location"],
                    str(eachrow["Line"]),
                ]

            # Note: using private property table._cells to avoid performance issue.  See
            # https://stackoverflow.com/a/69105798/316578
            col_index = 0
            for cell, cell_text in zip(table._cells, cells_text):
                cell.text = cell_text
                cell.width = widths[col_index]
                col_index = col_index + 1 if col_index < 2 else 0
        else:
            document.add_heading(f"Severity : {severity}", level=2)
            document.add_paragraph("None", style="List Bullet")



"""
SARIF file filtering functionality.
"""

import os
import re
from typing import Optional, List

import copy
import jsonpath_ng.ext
import yaml

from sarif import sarif_file_utils
from sarif.filter.filter_stats import FilterStats, load_filter_stats_from_json

# Commonly used properties can be specified using shortcuts
# instead of full JSON path
FILTER_SHORTCUTS = {
    "author": "properties.blame.author",
    "author-mail": "properties.blame.author-mail",
    "committer": "properties.blame.committer",
    "committer-mail": "properties.blame.committer-mail",
    "location": "locations[*].physicalLocation.artifactLocation.uri",
    "rule": "ruleId",
    "suppression": "suppressions[*].kind",
}

# Some properties can have specific shortcuts to make it easier to write filters
# For example a file location can be specified using wildcards
FIELDS_REGEX_SHORTCUTS = {"uri": {"**": ".*", "*": "[^/]*", "?": "."}}

# Default configuration for all filters
DEFAULT_CONFIGURATION = {
    "default-include": True,
    "check-line-number": True,
}


def _get_filter_function(filter_spec):
    """Return a filter function for the given specification."""
    if filter_spec:
        filter_len = len(filter_spec)
        if filter_len > 2 and filter_spec.startswith("/") and filter_spec.endswith("/"):
            regex = filter_spec[1:-1]
            return lambda value: re.search(regex, value, re.IGNORECASE)
        substring = filter_spec
        # substring can be empty, in this case "in" returns true
        # and only existence of the property checked.
        return lambda value: substring in value
    return lambda value: True


def _convert_glob_to_regex(property_name, property_value_spec):
    # skip if property_value_spec is a regex
    if property_value_spec and not (
        property_value_spec.startswith("/") and property_value_spec.endswith("/")
    ):
        # get last component of property name
        last_component = property_name.split(".")[-1]
        if last_component in FIELDS_REGEX_SHORTCUTS:
            shortcuts = FIELDS_REGEX_SHORTCUTS[last_component]
            regex = re.compile("|".join(map(re.escape, shortcuts.keys())))
            property_value_spec = regex.sub(
                lambda match: shortcuts[match.group(0)], property_value_spec
            )

            return f"/{property_value_spec}/"
    return property_value_spec


class PropertyFilter:
    """
    Class that represents a filter term ready for efficient use.
    """

    def __init__(self, prop_path, prop_value_spec, global_configuration):
        """
        Compile a filter property.  See README for the filter spec format.

        :param prop_path: JsonPath or preset.
        :param prop_value_spec: Value spec.
        :param global_configuration: Global configuration of the filter.
        """
        self.prop_path = prop_path
        resolved_prop_path = FILTER_SHORTCUTS.get(prop_path, prop_path)
        self.jsonpath_expr = jsonpath_ng.ext.parse(resolved_prop_path)

        # if prop_value_spec is a dict, update filter configuration from it
        if isinstance(prop_value_spec, dict):
            self.filter_configuration = copy.deepcopy(global_configuration)
            for config_key, config_value in prop_value_spec.items():
                if config_key != "value":
                    self.filter_configuration[config_key] = config_value
            # actual value for the filter is in "value" key
            prop_value_spec = prop_value_spec.get("value", "")
        else:
            self.filter_configuration = global_configuration
        value_spec = _convert_glob_to_regex(resolved_prop_path, prop_value_spec)
        self.filter_function = _get_filter_function(value_spec)


class MultiPropertyFilter:
    """
    Class representing a list of PropertyFilter objects.

    These are combined using AND to filter results.
    """

    def __init__(self, filter_spec: List[dict], global_filter_configuration: dict):
        """
        Initialise from a filter spec.

        See README for filter spec format.  It's a list of property paths and values to be
        combined with AND to form a filter.
        """
        self.filter_spec = filter_spec
        self.and_terms = [
            PropertyFilter(prop_path, prop_value_spec, global_filter_configuration)
            for prop_path, prop_value_spec in filter_spec.items()
        ]


def _compile_filters(
    filters: List[dict], global_filter_configuration: dict
) -> List[MultiPropertyFilter]:
    return [
        MultiPropertyFilter(filter_spec, global_filter_configuration)
        for filter_spec in filters
        if filter_spec
    ]


class GeneralFilter:
    """
    Class that implements filtering.
    """

    def __init__(self):
        self.filter_stats = None
        self.include_filters = {}
        self.apply_inclusion_filter = False
        self.exclude_filters = {}
        self.apply_exclusion_filter = False
        self.configuration = copy.deepcopy(DEFAULT_CONFIGURATION)

    def init_filter(
        self, filter_description, configuration, include_filters, exclude_filters
    ):
        """
        Initialise the filter with the given filter patterns.
        """
        self.filter_stats = FilterStats(filter_description)
        self.configuration.update(configuration)
        self.include_filters = _compile_filters(include_filters, self.configuration)
        self.apply_inclusion_filter = len(include_filters) > 0
        self.exclude_filters = _compile_filters(exclude_filters, self.configuration)
        self.apply_exclusion_filter = len(exclude_filters) > 0

    def rehydrate_filter_stats(self, dehydrated_filter_stats, filter_datetime):
        """
        Restore filter stats from the SARIF file directly,
        where they were recorded when the filter was previously run.

        Note that if init_filter is called,
        these rehydrated stats are discarded.
        """
        self.filter_stats = load_filter_stats_from_json(dehydrated_filter_stats)
        self.filter_stats.filter_datetime = filter_datetime

    def _zero_counts(self):
        if self.filter_stats:
            self.filter_stats.reset_counters()

    def _filter_append(self, filtered_results: List[dict], result: dict):
        # Remove any existing filter log on the result
        result.setdefault("properties", {}).pop("filtered", None)

        if self.apply_inclusion_filter:
            included_stats = self._filter_result(result, self.include_filters)
            if not included_stats["matchedFilter"]:
                # Result is excluded by dint of not being included
                self.filter_stats.filtered_out_result_count += 1
                return
        else:
            # no inclusion filters, mark the result as included so far
            included_stats = {"state": "included", "matchedFilter": []}

        if self.apply_exclusion_filter:
            excluded_stats = self._filter_result(result, self.exclude_filters)
            if excluded_stats["matchedFilter"]:
                self.filter_stats.filtered_out_result_count += 1
                return

        included_state = included_stats["state"]
        if included_state == "included":
            self.filter_stats.filtered_in_result_count += 1
        elif included_state == "noLineNumber":
            self.filter_stats.unconvincing_line_number_count += 1
        else:
            self.filter_stats.missing_property_count += 1
        included_stats["filter"] = self.filter_stats.filter_description
        result["properties"]["filtered"] = included_stats

        filtered_results.append(result)

    def _filter_result(self, result: dict, filters: List[MultiPropertyFilter]) -> dict:
        matched_filters = []
        warnings = []
        (_file_path, line_number) = sarif_file_utils.read_result_location(result)
        unconvincing_line_number = line_number == "1" or not line_number
        default_include_noprop = False

        if filters:
            # filters contain rules which treated as OR.
            # if any rule matches, the record is selected.
            for mpf in filters:
                # filter_spec contains rules which treated as AND.
                # all rules must match to select the record.
                matched = True
                for property_filter in mpf.and_terms:
                    if (
                        property_filter.filter_configuration.get(
                            "check-line-number", True
                        )
                        and unconvincing_line_number
                    ):
                        warnings.append(
                            f"Field '{property_filter.prop_path}' not checked due to "
                            "missing line number information"
                        )
                        continue
                    found_results = property_filter.jsonpath_expr.find(result)
                    if found_results:
                        value = found_results[0].value
                        if property_filter.filter_function(value):
                            continue
                    else:
                        # property to filter on is not found, or skipped due to invalid line number.
                        # if "default-include" is true, include the "result" with a warning.
                        if property_filter.filter_configuration.get(
                            "default-include", True
                        ):
                            warnings.append(
                                f"Field '{property_filter.prop_path}' is missing but "
                                "the result included as default-include is true"
                            )
                            default_include_noprop = True
                            continue
                    matched = False
                    break
                if matched:
                    matched_filters.append(mpf.filter_spec)
                    break

        stats = {
            "state": "included",
            "matchedFilter": matched_filters,
        }

        if warnings:
            stats.update(
                {
                    "state": "noProperty" if default_include_noprop else "noLineNumber",
                    "warnings": warnings,
                }
            )

        return stats

    def filter_results(self, results: List[dict]) -> List[dict]:
        """
        Apply this filter to a list of results,
        return the results that pass the filter
        and as a side-effect, update the filter stats.
        """
        if self.apply_inclusion_filter or self.apply_exclusion_filter:
            self._zero_counts()
            ret = []
            for result in results:
                self._filter_append(ret, result)
            return ret
        # No inclusion or exclusion patterns
        return results

    def get_filter_stats(self) -> Optional[FilterStats]:
        """
        Get the statistics from running this filter.
        """
        return self.filter_stats


def load_filter_file(file_path):
    """
    Load a YAML filter file, return the filter description and the filters.
    """
    try:
        file_name = os.path.basename(file_path)
        with open(file_path, encoding="utf-8") as file_in:
            yaml_content = yaml.safe_load(file_in)
            filter_description = yaml_content.get("description", file_name)
            configuration = yaml_content.get("configuration", {})
            include_filters = yaml_content.get("include", {})
            exclude_filters = yaml_content.get("exclude", {})
    except yaml.YAMLError as error:
        raise IOError(f"Cannot read filter file {file_path}") from error
    return filter_description, configuration, include_filters, exclude_filters



"""
Code for `sarif diff` command.
"""

import json
import sys
from typing import Dict

from sarif import sarif_file


def _occurrences(occurrence_count):
    return (
        "1 occurrence" if occurrence_count == 1 else f"{occurrence_count} occurrences"
    )


def _signed_change(difference):
    return str(difference) if difference < 0 else f"+{difference}"


def _record_to_location_tuple(record) -> str:
    return (record["Location"], record["Line"])


def print_diff(
    old_sarif: sarif_file.SarifFileSet,
    new_sarif: sarif_file.SarifFileSet,
    output,
    check_level=None,
) -> int:
    """
    Generate a diff of the issues from the SARIF files and write it to stdout
    or a file if specified.
    :param old_sarif: corresponds to the old files.
    :param new_sarif: corresponds to the new files.
    :return: number of increased severities, or 0 if nothing has worsened.
    """
    diff = _calc_diff(old_sarif, new_sarif)
    if output:
        print("writing diff to", output)
        with open(output, "w", encoding="utf-8") as output_file:
            json.dump(diff, output_file, indent=4)
    else:
        for severity in sarif_file.SARIF_SEVERITIES_WITH_NONE:
            if severity not in diff:
                continue
            if diff[severity]["codes"]:
                print(
                    severity,
                    "level:",
                    _signed_change(diff[severity]["+"]),
                    _signed_change(-diff[severity]["-"]),
                )
                for issue_key, code_info in diff[severity]["codes"].items():
                    (old_count, new_count, new_locations) = (
                        code_info["<"],
                        code_info[">"],
                        code_info.get("+@", []),
                    )
                    if old_count == 0:
                        print(f'  New issue "{issue_key}" ({_occurrences(new_count)})')
                    elif new_count == 0:
                        print(f'  Eliminated issue "{issue_key}"')
                    else:
                        print(
                            f"  Number of occurrences {old_count} -> {new_count}",
                            f'({_signed_change(new_count - old_count)}) for issue "{issue_key}"',
                        )
                    if new_locations:
                        # Print the top 3 new locations
                        for record in new_locations[0:3]:
                            (location, line) = _record_to_location_tuple(record)
                            print(f"    {location}:{line}")
                        if len(new_locations) > 3:
                            print("    ...")
            else:
                print(severity, "level: +0 -0 no changes")
        print(
            "all levels:",
            _signed_change(diff["all"]["+"]),
            _signed_change(-diff["all"]["-"]),
        )
    filter_stats = old_sarif.get_filter_stats()
    if filter_stats:
        print(f"  'Before' results were filtered by {filter_stats}")
    filter_stats = new_sarif.get_filter_stats()
    if filter_stats:
        print(f"  'After' results were filtered by {filter_stats}")
    ret = 0
    if check_level:
        for severity in sarif_file.SARIF_SEVERITIES_WITH_NONE:
            ret += diff.get(severity, {}).get("+", 0)
            if severity == check_level:
                break
    if ret > 0:
        sys.stderr.write(
            f"Check: exiting with return code {ret} due to increase in issues at or above {check_level} severity\n"
        )
    return ret


def _find_new_occurrences(new_records, old_records):
    # Note: this is O(n²) complexity where n is the number of occurrences of this issue type,
    # so could be slow when there are a large number of occurrences.
    old_occurrences = old_records
    new_occurrences_new_locations = []
    new_occurrences_new_lines = []
    for new_record in new_records:
        (new_location, new_line) = (True, True)
        for old_record in old_occurrences:
            if old_record["Location"] == new_record["Location"]:
                new_location = False
                if old_record["Line"] == new_record["Line"]:
                    new_line = False
                    break
        if new_location:
            if new_record not in new_occurrences_new_locations:
                new_occurrences_new_locations.append(new_record)
        elif new_line:
            if new_record not in new_occurrences_new_lines:
                new_occurrences_new_lines.append(new_record)

    return sorted(
        new_occurrences_new_locations, key=_record_to_location_tuple
    ) + sorted(new_occurrences_new_lines, key=_record_to_location_tuple)


def _calc_diff(
    old_sarif: sarif_file.SarifFileSet, new_sarif: sarif_file.SarifFileSet
) -> Dict:
    """
    Generate a diff of the issues from the SARIF files.
    old_sarif corresponds to the old files.
    new_sarif corresponds to the new files.
    Return dict has keys "error", "warning", "note", "none" (if present) and "all".
    """
    ret = {"all": {"+": 0, "-": 0}}
    old_report = old_sarif.get_report()
    new_report = new_sarif.get_report()
    # Include `none` in the list of severities if there are any `none` records in either the old
    # or new report.
    severities = (
        old_report.get_severities()
        if old_report.any_none_severities()
        else new_report.get_severities()
    )
    for severity in severities:
        old_histogram = old_report.get_issue_type_histogram_for_severity(severity)
        new_histogram = new_report.get_issue_type_histogram_for_severity(severity)
        ret[severity] = {"+": 0, "-": 0, "codes": {}}
        if old_histogram != new_histogram:
            for issue_key, count in new_histogram.items():
                old_count = old_histogram.pop(issue_key, 0)
                if old_count != count:
                    ret[severity]["codes"][issue_key] = {"<": old_count, ">": count}
                    if old_count == 0:
                        ret[severity]["+"] += 1
                    new_occurrences = _find_new_occurrences(
                        new_report.get_issues_grouped_by_type_for_severity(
                            severity
                        ).get(issue_key, []),
                        old_report.get_issues_grouped_by_type_for_severity(
                            severity
                        ).get(issue_key, []),
                    )
                    if new_occurrences:
                        ret[severity]["codes"][issue_key]["+@"] = [
                            {"Location": r["Location"], "Line": r["Line"]}
                            for r in new_occurrences
                        ]
            for issue_key, old_count in old_histogram.items():
                ret[severity]["codes"][issue_key] = {"<": old_count, ">": 0}
                ret[severity]["-"] += 1
        ret["all"]["+"] += ret[severity]["+"]
        ret["all"]["-"] += ret[severity]["-"]
    return ret



"""
Code for `sarif blame` command.
"""

import json
import os
import subprocess
import sys
import urllib.parse
import urllib.request

from sarif.sarif_file import SarifFileSet


def enhance_with_blame(
    input_files: SarifFileSet, repo_path: str, output: str, output_multiple_files: bool
):
    """
    Enhance SARIF files with information from `git blame`.  The `git` command is run in the current
    directory, which must be a git repository containing the files at the paths specified in the
    input files.  Updated files are written to output_path if specified, otherwise to the current
    directory.
    """
    if not input_files:
        return
    if not os.path.isdir(repo_path):
        raise ValueError(f"No git repository directory found at {repo_path}")

    _enhance_with_blame(input_files, repo_path)

    for input_file in input_files:
        input_file_name = input_file.get_file_name()
        if any(
            "blame" in result.get("properties", {})
            for result in input_file.get_results()
        ):
            output_file = output
            if output_multiple_files:
                output_filename = (
                    input_file.get_file_name_without_extension()
                    + "_with_blame."
                    + input_file.get_file_name_extension()
                )
                output_file = os.path.join(output, output_filename)
            print(
                "Writing",
                output_file,
                "combining original SARIF from",
                input_file_name,
                "with git blame information",
            )
            with open(output_file, "w", encoding="utf-8") as file_out:
                json.dump(input_file.data, file_out)
        else:
            sys.stderr.write(
                f"WARNING: did not find any git blame information for {input_file_name}\n"
            )


def _enhance_with_blame(input_files, repo_path):
    """
    Run `git blame --porcelain` for each file path listed in input_files.
    Then enhance the results in error_list by adding a "blame" property including "hash", "author"
    and "timestamp".
    Porcelain format is used for parseability and stability.  See documentation at
    https://git-scm.com/docs/git-blame#_the_porcelain_format.
    """
    files_to_blame = set(item["Location"] for item in input_files.get_records())
    file_count = len(files_to_blame)
    print(
        "Running `git blame --porcelain` on",
        "one file" if file_count == 1 else f"{file_count} files",
        "in",
        repo_path,
    )
    file_blame_info = _run_git_blame_on_files(files_to_blame, repo_path)

    # Now join up blame output with result list
    blame_info_count = 0
    item_count = 0
    for result, record in zip(input_files.get_results(), input_files.get_records()):
        item_count += 1
        file_path = record["Location"]
        if file_path in file_blame_info:
            blame_info = file_blame_info[file_path]
            # raw_line can be None if no line number information was included in the SARIF result.
            raw_line = record["Line"]
            if raw_line:
                line_no = str(raw_line)
                if line_no in blame_info["line_to_commit"]:
                    commit_hash = blame_info["line_to_commit"][line_no]
                    commit = blame_info["commits"][commit_hash]
                    # Add blame information to the SARIF Property Bag of the result
                    result.setdefault("properties", {})["blame"] = commit
                    blame_info_count += 1
    print(f"Found blame information for {blame_info_count} of {item_count} results")


def _make_path_git_compatible(file_path):
    try:
        path_as_url = urllib.parse.urlparse(file_path)
        if path_as_url.scheme == "file":
            return urllib.request.url2pathname(path_as_url.path)
        return file_path
    except ValueError:
        return file_path


def _run_git_blame_on_files(files_to_blame, repo_path):
    file_blame_info = {}
    for file_path in files_to_blame:
        cmd = ["git", "blame", "--porcelain", _make_path_git_compatible(file_path)]
        with subprocess.Popen(cmd, stdout=subprocess.PIPE, cwd=repo_path) as proc:
            blame_info = {"commits": {}, "line_to_commit": {}}
            file_blame_info[file_path] = blame_info
            reading_commit_headers = False
            for line_bytes in proc.stdout.readlines():
                # Convert byte sequence to string and remove trailing LF
                line_string = line_bytes.decode("utf-8")[:-1]
                # Now parse output from git blame --porcelain
                if reading_commit_headers:
                    if line_string.startswith("\t"):
                        reading_commit_headers = False
                        # Ignore line contents = source code
                    elif " " in line_string:
                        space_pos = line_string.index(" ")
                        key = line_string[0:space_pos]
                        value = line_string[space_pos + 1 :].strip()
                        blame_info["commits"][commit_hash][key] = value
                    else:
                        # e.g. "boundary"
                        key = line_string
                        blame_info["commits"][commit_hash][key] = True
                else:
                    commit_line_info = line_string.split(" ")
                    commit_hash = commit_line_info[0]
                    commit_line = commit_line_info[2]
                    blame_info["commits"].setdefault(commit_hash, {})
                    blame_info["line_to_commit"][commit_line] = commit_hash
                    reading_commit_headers = True
            # Ensure process terminates
            proc.communicate()
            if proc.returncode:
                cmd_str = " ".join(cmd)
                sys.stderr.write(
                    f"WARNING: Command `{cmd_str} "
                    f"failed with exit code {proc.returncode} in {repo_path}\n"
                )
    return file_blame_info
